@@ -5,6 +5,7 @@ from functools import reduce
 from operator import mul
 from . import prune
 from enum import IntEnum
+from numbers import Number
 
 __all__ = ['PruningPlan', 'Dependency', 'DependencyGraph']
 
@@ -18,14 +19,19 @@ class OPTYPE(IntEnum):
     BN = 1
     LINEAR = 2
     PRELU = 3
+    GROUP_CONV=4
 
-    CONCAT=4
-    ELEMENTWISE=5
-
+    CONCAT=5
+    SPLIT=6
+    CUSTOMIZED=7
+    ELEMENTWISE=8
 
 def _get_module_type(module):
     if isinstance( module, TORCH_CONV ):
-        return OPTYPE.CONV
+        if module.groups>1:
+            return OPTYPE.GROUP_CONV
+        else:
+            return OPTYPE.CONV
     elif isinstance( module, TORCH_BATCHNORM ):
         return OPTYPE.BN
     elif isinstance( module, TORCH_PRELU ):
@@ -34,11 +40,15 @@ def _get_module_type(module):
         return OPTYPE.LINEAR
     elif isinstance( module, _ConcatOp ):
         return OPTYPE.CONCAT
+    elif isinstance( module, _SplitOp):
+        return OPTYPE.SPLIT
+    elif isinstance(module, _CustomizedOp):
+        return OPTYPE.CUSTOMIZED
     else:
         return OPTYPE.ELEMENTWISE
 
-def _get_node_channel(node):
-    if node.type==OPTYPE.CONV:
+def _get_node_out_channel(node):
+    if node.type==OPTYPE.CONV or node.type==OPTYPE.GROUP_CONV:
         return node.module.out_channels
     elif node.type==OPTYPE.BN:
         return node.module.num_features
@@ -49,6 +59,25 @@ def _get_node_channel(node):
             return None
         else:
             return node.module.num_parameters
+    elif node.type==OPTYPE.CUSTOMIZED:
+        return node.customized_op_fn['get_out_ch_fn'](node.module)
+    else:
+        return None
+
+def _get_node_in_channel(node):
+    if node.type==OPTYPE.CONV or node.type==OPTYPE.GROUP_CONV:
+        return node.module.in_channels
+    elif node.type==OPTYPE.BN:
+        return node.module.num_features
+    elif node.type==OPTYPE.LINEAR:
+        return node.module.in_features
+    elif node.type==OPTYPE.PRELU:
+        if node.module.num_parameters==1:
+            return None
+        else:
+            return node.module.num_parameters
+    elif node.type==OPTYPE.CUSTOMIZED:
+        return node.customized_op_fn['get_in_ch_fn'](node.module)
     else:
         return None
 
@@ -56,9 +85,19 @@ def _get_node_channel(node):
 def _prune_concat(layer, *args, **kargs):
     return layer, 0
 
+def _prune_split(layer, *args, **kargs):
+    return layer, 0
+
 def _prune_elementwise_op(layer, *args, **kargs):
     return layer, 0
 
+class _CustomizedOp(nn.Module):
+    def __init__(self, op_class):
+        self.op_cls = op_class
+
+    def __repr__(self):
+        return "CustomizedOp(%s)"%(str(self.op_cls))
+
 # Dummy module
 class _ConcatOp(nn.Module):
     def __init__(self):
@@ -68,6 +107,14 @@ class _ConcatOp(nn.Module):
     def __repr__(self):
         return "_ConcatOp(%s)"%(self.offsets)
 
+class _SplitOp(nn.Module):
+    def __init__(self):
+        super(_SplitOp, self).__init__()
+        self.offsets = None
+        
+    def __repr__(self):
+        return "_SplitOp(%s)"%(self.offsets)
+
 class _ElementWiseOp(nn.Module):
     def __init__(self):
         super(_ElementWiseOp, self).__init__()
@@ -75,11 +122,7 @@ class _ElementWiseOp(nn.Module):
     def __repr__(self):
         return "_ElementWiseOp()"
 
-class _Inputs(nn.Module):
-    def __init__(self):
-        super(_Inputs, self).__init__()
-
-class _StrideIndexTransform(object):
+class _FlattenIndexTransform(object):
     def __init__(self, stride=1, reverse=False):
         self._stride = stride
         self.reverse = reverse
@@ -95,7 +138,8 @@ class _StrideIndexTransform(object):
                 new_idxs.extend(list( range(i*self._stride, (i+1)*self._stride)))
         return new_idxs
 
-class _OffsetIndexTransform(object):
+
+class _ConcatIndexTransform(object):
     def __init__(self, offset, reverse=False):
         self.offset = offset
         self.reverse = reverse
@@ -106,6 +150,19 @@ class _OffsetIndexTransform(object):
         else:
             new_idxs = [i+self.offset[0] for i in idxs]
         return new_idxs
+
+
+class _SplitIndexTransform(object):
+    def __init__(self, offset, reverse=False):
+        self.offset = offset
+        self.reverse = reverse
+
+    def __call__(self, idxs):
+        if self.reverse==True:
+            new_idxs = [i+self.offset[0] for i in idxs ]
+        else:
+            new_idxs = [i-self.offset[0] for i in idxs if (i>=self.offset[0] and i<self.offset[1])]
+        return new_idxs
         
 class Node(object):
     def __init__(self, module, grad_fn, node_name=None):
@@ -122,10 +179,12 @@ class Node(object):
         return "%s (%s)"%(self._node_name, str(self.module)) if self._node_name is not None else str(self.module)
 
     def add_input(self, node):
-        self.inputs.append( node )
+        if node not in self.inputs:
+            self.inputs.append( node )
     
     def add_output(self, node):
-        self.outputs.append( node )
+        if node not in self.outputs:
+            self.outputs.append( node )
 
     def __repr__(self):
         return "<Node: (%s, %s)>"%( self.node_name, self.grad_fn )
@@ -238,42 +297,81 @@ class PruningPlan(object):
 
 class DependencyGraph(object):
 
-    PRUNABLE_MODULES = ( nn.modules.conv._ConvNd, nn.modules.batchnorm._BatchNorm, nn.Linear, nn.PReLU )
-    ELEMENTWISE_GRAD_FN = ('AddBackward', 'SubBackward', 'MulBackward', 'DivBackward', 'PowBackward', )
-    CONCAT_GRAD_FN = ('CatBackward',)
+    PRUNABLE_MODULES = [ nn.modules.conv._ConvNd, nn.modules.batchnorm._BatchNorm, nn.Linear, nn.PReLU ]
     
-    HANDLER = {                         # prune in_channel          # prune out_channel
-                OPTYPE.CONV          :  (prune.prune_related_conv,   prune.prune_conv),
-                OPTYPE.BN            :  (prune.prune_batchnorm,      prune.prune_batchnorm),
-                OPTYPE.PRELU         :  (prune.prune_prelu,          prune.prune_prelu),
-                OPTYPE.LINEAR        :  (prune.prune_related_linear, prune.prune_linear),
-                OPTYPE.CONCAT        :  (_prune_concat,              _prune_concat),
-                OPTYPE.ELEMENTWISE   :  (_prune_elementwise_op,      _prune_elementwise_op),
-               }
+    HANDLER = {    # pruning function that changes 1. in_channel           2. out_channel
+                OPTYPE.CONV                 :  (prune.prune_related_conv,   prune.prune_conv),
+                OPTYPE.BN                   :  (prune.prune_batchnorm,      prune.prune_batchnorm),
+                OPTYPE.PRELU                :  (prune.prune_prelu,          prune.prune_prelu),
+                OPTYPE.LINEAR               :  (prune.prune_related_linear, prune.prune_linear),
+                OPTYPE.GROUP_CONV           :  (prune.prune_group_conv,     prune.prune_group_conv),
+                OPTYPE.CONCAT               :  (_prune_concat,              _prune_concat),
+                OPTYPE.SPLIT                :  (_prune_split,               _prune_split),
+                OPTYPE.ELEMENTWISE          :  (_prune_elementwise_op,      _prune_elementwise_op),
+                OPTYPE.CUSTOMIZED           :  (None, None), # placeholder
+            }
     OUTPUT_NODE_RULES = {}
     INPUT_NODE_RULES = {}
     for t1 in HANDLER.keys():
         for t2 in HANDLER.keys():
             OUTPUT_NODE_RULES[ (t1, t2) ] = (HANDLER[t1][1], HANDLER[t2][0]) # change in_channels of output layer
-            INPUT_NODE_RULES[ (t1, t2) ] = (HANDLER[t1][0], HANDLER[t2][1]) # change out_channels of input layer
+            INPUT_NODE_RULES[ (t1, t2) ] = (HANDLER[t1][0], HANDLER[t2][1])  # change out_channels of input layer
+    CUSTOMIZED_OP_FN = {}
 
-    def build_dependency( self, model, example_inputs, get_output_fn=None ):
+    def build_dependency( self, 
+        model:torch.nn.Module, 
+        example_inputs: typing.Union[torch.Tensor, typing.Sequence],
+        output_transform:typing.Callable=None, 
+        verbose:bool=True ):
+        """ Build a dependency graph through forwarding.
+
+        Parameters:
+            model (class): the model to be pruned.
+            example_inputs (torch.Tensor or List): dummy inputs for the model.
+            output_transform (Callable): A function to transform network outputs.
+            verbose (Callable): verbose mode.
+        """
+
+        self.verbose = verbose
         # get module name
         self._module_to_name = { module: name for (name, module) in model.named_modules() }
         # build dependency graph
-        self.module_to_node = self._obtain_forward_graph( model, example_inputs, get_output_fn=get_output_fn )
+        self.module_to_node = self._obtain_forward_graph( model, example_inputs, output_transform=output_transform )
         self._build_dependency(self.module_to_node)
         self.update_index()
         return self
 
-    def update_index( self ):
-        for module, node in self.module_to_node.items():
-            if node.type==OPTYPE.LINEAR:
-                self._set_fc_index_transform( node )
-            if node.type==OPTYPE.CONCAT:
-                self._set_concat_index_transform(node)
+    def register_customized_layer(self, layer_type, in_ch_pruning_fn, out_ch_pruning_fn, get_in_ch_fn, get_out_ch_fn):
+        """ Register a customized layer for pruning.
+
+        Parameters:
+            layer_type (class): the type of layer
+            in_ch_pruning_fn (Callable): A function to prune channels/dimensions of input tensor
+            out_ch_pruning_fn (Callable): A function to prune channels/dimensions of output tensor
+            get_in_ch_fn (Callable): estimate the n_channel of layer input. Return None if the layer does not change tensor shape.
+            get_out_ch_fn (Callable):estimate the n_channel of layer output. Return None if the layer does not change tensor shape.
+        """
+        self.CUSTOMIZED_OP_FN[layer_type] = {
+            "in_ch_pruning_fn": in_ch_pruning_fn, 
+            "out_ch_pruning_fn": out_ch_pruning_fn, 
+            "get_in_ch_fn": get_in_ch_fn, 
+            "get_out_ch_fn": get_out_ch_fn,
+        }
+        self.PRUNABLE_MODULES.append( layer_type )
+ 
+    def get_pruning_plan(self, module: nn.Module, pruning_fn: typing.Callable, idxs: typing.Union[list, tuple]): 
+        """ Get a pruning plan from the dependency graph, according to user's pruning operations. 
+
+        Parameters:
+            module (nn.Module): the module to be pruned.
+            pruning_fn (Callable): the pruning function.
+            idxs (list or tuple): the indices of paramters to be pruned.
+        """
+        if isinstance(module, TORCH_CONV) and module.groups>1:
+            pruning_fn = prune.prune_group_conv
+        if isinstance(idxs, Number):
+            idxs = [idxs]
 
-    def get_pruning_plan(self, module, pruning_fn, idxs): 
         self.update_index()
         plan = PruningPlan()
         #  the user pruning operation
@@ -283,7 +381,6 @@ class DependencyGraph(object):
         visited = set()
         def _fix_denpendency_graph(node, fn, indices):
             visited.add( node )
-
             for dep in node.dependencies:
                 if dep.is_triggered_by(fn): #and dep.broken_node not in visited:
                     if dep.index_transform is not None:
@@ -305,93 +402,137 @@ class DependencyGraph(object):
         for dep, idxs in plan.plan:
             merged_plan.add_plan_and_merge( dep, idxs )
         return merged_plan
-    
+
     def _build_dependency(self, module_to_node):
         for module, node in module_to_node.items():
             for in_node in node.inputs:
                 in_node_rule = self.INPUT_NODE_RULES.get( (node.type, in_node.type), None )
                 if in_node_rule is not None:
-                    dep = Dependency( trigger=in_node_rule[0], handler=in_node_rule[1], broken_node=in_node)
+                    trigger = in_node_rule[0]
+                    handler = in_node_rule[1]
+                    if trigger is None:
+                        trigger = self.CUSTOMIZED_OP_FN[type(node.module)]['in_ch_pruning_fn']
+                    if handler is None:
+                        handler = self.CUSTOMIZED_OP_FN[type(in_node.module)]['out_ch_pruning_fn']
+                    dep = Dependency( trigger=trigger, handler=handler, broken_node=in_node)
                     node.dependencies.append( dep )
 
             for out_node in node.outputs:
                 out_node_rule = self.OUTPUT_NODE_RULES.get( (node.type, out_node.type), None )
                 if out_node_rule is not None:
-                    dep = Dependency( trigger=out_node_rule[0], handler=out_node_rule[1], broken_node=out_node)
+                    trigger = out_node_rule[0]
+                    handler = out_node_rule[1]
+                    if trigger is None:
+                        trigger = self.CUSTOMIZED_OP_FN[type(node.module)]['out_ch_pruning_fn']
+                    if handler is None:
+                        handler = self.CUSTOMIZED_OP_FN[type(out_node.module)]['in_ch_pruning_fn']
+                    dep = Dependency( trigger=trigger, handler=handler, broken_node=out_node)
                     node.dependencies.append( dep )
     
-    def _obtain_forward_graph(self, model, example_inputs, get_output_fn):
+    def _obtain_forward_graph(self, model, example_inputs, output_transform):
         #module_to_node = { m: Node( m ) for m in model.modules() if isinstance( m, self.PRUNABLE_MODULES ) }
         model.eval().cpu()
         # Get grad_fn from prunable modules
         grad_fn_to_module = {}
-        def _record_module_grad_fn(module, inputs, outputs):
-            grad_fn_to_module[outputs.grad_fn] = module
 
-        hooks = [m.register_forward_hook(_record_module_grad_fn) for m in model.modules() if isinstance( m, self.PRUNABLE_MODULES ) ]
-        out = model(example_inputs)
+        visited = {}
+        def _record_module_grad_fn(module, inputs, outputs):
+            if module not in visited:
+                visited[module] = 1
+            else:
+                visited[module] += 1
+            grad_fn_to_module[outputs.grad_fn] = module
+        
+        hooks = [m.register_forward_hook(_record_module_grad_fn) for m in model.modules() if isinstance( m, tuple(self.PRUNABLE_MODULES )) ]
+        
+        if isinstance(example_inputs, (tuple, list)):
+            out = model(*example_inputs)
+        elif isinstance(example_inputs, dict):
+            out = model(**example_inputs)
+        elif isinstance(example_inputs, torch.Tensor):
+            out = model(example_inputs)
+
         for hook in hooks:
             hook.remove()
-        grad_fn_from_prunable_module = list( grad_fn_to_module.keys() )
-
+        reused = [ m for (m, count) in visited.items() if count>1 ]
         # create nodes and dummy modules
-        visited = set()
         module_to_node = {}
-
         def _build_graph(grad_fn):
-            module = grad_fn_to_module.get( grad_fn, None )
-            if module is not None and module in module_to_node:
+            module = grad_fn_to_module.get( grad_fn, None )   
+            if module is not None and module in module_to_node and module not in reused:
                 return module_to_node[module]
             
-            if grad_fn in grad_fn_from_prunable_module: # prunable ops
-                module = grad_fn_to_module[ grad_fn ]
-            else: # create dummy modules
-                if 'catbackward' in grad_fn.name().lower(): # concat op
+            if module is None:
+                if not hasattr(grad_fn, 'name'):
+                    module = _ElementWiseOp() # skip customized modules
+                    if self.verbose:
+                        print("[Warning] Unrecognized operation: %s. It will be treated as element-wise op"%( str(grad_fn) ))
+                elif 'catbackward' in grad_fn.name().lower(): # concat op
                     module = _ConcatOp()
+                elif 'splitbackward' in grad_fn.name().lower():
+                    module = _SplitOp()
                 else:
                     module = _ElementWiseOp()   # All other ops are treated as element-wise ops
                 grad_fn_to_module[ grad_fn ] = module # record grad_fn
 
-            node = Node( module, grad_fn, self._module_to_name.get( module, None ) )
-            module_to_node[ module ] = node
+            if module not in module_to_node:
+                node = Node( module, grad_fn, self._module_to_name.get( module, None ) )
+                if type(module) in self.CUSTOMIZED_OP_FN.keys(): # mark it as a customized OP
+                    node.type = OPTYPE.CUSTOMIZED
+                    node.customized_op_fn = self.CUSTOMIZED_OP_FN[type(module)]
+                module_to_node[ module ] = node
+            else:
+                node = module_to_node[module]
 
             if hasattr(grad_fn, 'next_functions'):
                 for f in grad_fn.next_functions:
-                    if f[0] is not None and 'accumulategrad' not in f[0].name().lower():
+                    if f[0] is not None:
+                        if hasattr( f[0], 'name' ) and 'accumulategrad' in f[0].name().lower(): # skip leaf variables
+                            continue
                         input_node = _build_graph(f[0])
-                        # connect nodes
                         node.add_input( input_node )
                         input_node.add_output( node )
-            #print(grad_fn, node)
             return node
         
-        if get_output_fn is not None:
-            out = get_output_fn(out)
-            
-        if isinstance(out, (list, tuple) ):
-            for o in out:
-                _build_graph( o.grad_fn )
-        else:
-            _build_graph( out.grad_fn )
+        if output_transform is not None:
+            out = output_transform(out)
+        
+        #if isinstance(out, (list, tuple) ):
+        #    for o in out:
+        #        _build_graph( o.grad_fn )
+        #else:
+        #    _build_graph( out.grad_fn )
+        for o in flatten_as_list(out):
+             _build_graph( o.grad_fn )
         return module_to_node
 
+    def update_index( self ):
+        for module, node in self.module_to_node.items():
+            if node.type==OPTYPE.LINEAR:
+                self._set_fc_index_transform( node )
+            if node.type==OPTYPE.CONCAT:
+                self._set_concat_index_transform(node)
+            if node.type==OPTYPE.SPLIT:
+                self._set_split_index_transform(node)
+
     def _set_fc_index_transform(self, fc_node: Node):
         if fc_node.type != OPTYPE.LINEAR:
             return
-        
         visited = set()
         fc_in_features = fc_node.module.in_features
-        feature_channels = _get_input_channels(fc_node.inputs[0])
+        feature_channels = _get_out_channels_of_in_node(fc_node.inputs[0])
+        if feature_channels<=0: # the first layer: https://github.com/VainF/Torch-Pruning/issues/21
+            return
         stride = fc_in_features // feature_channels
         if stride>1:
             for in_node in fc_node.inputs:
                 for dep in fc_node.dependencies:
                     if dep.broken_node==in_node:
-                        dep.index_transform = _StrideIndexTransform( stride=stride, reverse=True )
+                        dep.index_transform = _FlattenIndexTransform( stride=stride, reverse=True )
                 
                 for dep in in_node.dependencies:
                     if dep.broken_node == fc_node:
-                        dep.index_transform = _StrideIndexTransform( stride=stride, reverse=False )
+                        dep.index_transform = _FlattenIndexTransform( stride=stride, reverse=False )
 
     def _set_concat_index_transform(self, cat_node: Node):
         if cat_node.type != OPTYPE.CONCAT:
@@ -399,7 +540,7 @@ class DependencyGraph(object):
         
         chs = []
         for n in cat_node.inputs:
-            chs.append( _get_input_channels(n) )
+            chs.append( _get_out_channels_of_in_node(n) )
 
         offsets = [0]
         for ch in chs:
@@ -409,19 +550,68 @@ class DependencyGraph(object):
         for i, in_node in enumerate(cat_node.inputs):
             for dep in cat_node.dependencies:
                 if dep.broken_node == in_node:
-                    dep.index_transform = _OffsetIndexTransform( offset=offsets[i:i+2], reverse=True )
+                    dep.index_transform = _ConcatIndexTransform( offset=offsets[i:i+2], reverse=True )
 
             for dep in in_node.dependencies:
                 if dep.broken_node == cat_node:
-                    dep.index_transform = _OffsetIndexTransform( offset=offsets[i:i+2], reverse=False )
+                    dep.index_transform = _ConcatIndexTransform( offset=offsets[i:i+2], reverse=False )
 
-def _get_input_channels(node):
-    ch = _get_node_channel(node)
+    def _set_split_index_transform(self, split_node: Node):
+        if split_node.type != OPTYPE.SPLIT:
+            return
+        
+        chs = []
+        for n in split_node.outputs:
+            chs.append( _get_in_channels_of_out_node(n) )
+
+        offsets = [0]
+        for ch in chs:
+            offsets.append( offsets[-1]+ch )
+        split_node.module.offsets = offsets
+        for i, out_node in enumerate(split_node.outputs):
+            for dep in split_node.dependencies:
+                if dep.broken_node == out_node:
+                    dep.index_transform = _SplitIndexTransform( offset=offsets[i:i+2], reverse=False )
+
+            for dep in out_node.dependencies:
+                if dep.broken_node == split_node:
+                    dep.index_transform = _SplitIndexTransform( offset=offsets[i:i+2], reverse=True )
+
+def _get_out_channels_of_in_node(node):
+    ch = _get_node_out_channel(node)
     if ch is None:
         ch = 0
         for in_node in node.inputs:
             if node.type==OPTYPE.CONCAT:
-                ch+=_get_input_channels( in_node )
+                ch+=_get_out_channels_of_in_node( in_node )
             else:
-                ch=_get_input_channels( in_node )
-    return ch
\ No newline at end of file
+                ch=_get_out_channels_of_in_node( in_node )
+    return ch
+
+def _get_in_channels_of_out_node(node):
+    ch = _get_node_in_channel(node)
+    if ch is None:
+        ch = 0
+        for out_node in node.outputs:
+            if node.type==OPTYPE.SPLIT:
+                ch+=_get_in_channels_of_out_node( out_node )
+            else:
+                ch=_get_in_channels_of_out_node( out_node )
+    return ch
+
+def flatten_as_list(obj):
+    if isinstance(obj, torch.Tensor):
+        return [ obj ]
+    elif isinstance(obj, (list,tuple)):
+        flattened_list = []
+        for sub_obj in obj:
+            flattened_list.extend( flatten_as_list(sub_obj) )
+        return flattened_list
+    elif isinstance(obj, dict):
+        flattened_list = []
+        for sub_obj in obj.values():
+            flattened_list.extend( flatten_as_list(sub_obj) )
+        return flattened_list
+    else:
+        return obj
+    
\ No newline at end of file
