@@ -1,4 +1,5 @@
 from .dependency import TORCH_CONV, TORCH_BATCHNORM, TORCH_PRELU, TORCH_LINEAR
+import torch
 
 def count_prunable_params(module):
     if isinstance( module, ( TORCH_CONV, TORCH_LINEAR) ):
@@ -21,7 +22,7 @@ def count_prunable_params(module):
 
 def count_prunable_channels(module):
     if isinstance( module, TORCH_CONV ):
-        return module.out_channels
+        return module.weight.shape[0]
     elif isinstance( module, TORCH_LINEAR ):
         return module.out_features
     elif isinstance( module, TORCH_BATCHNORM ):
@@ -32,4 +33,8 @@ def count_prunable_channels(module):
         else:
             return len(module.weight)
     else:
-        return 0
\ No newline at end of file
+        return 0
+
+def count_params(module):
+    return sum([ p.numel() for p in module.parameters() ])
+
